package de.steamwar.core;

import com.google.common.base.Preconditions;
import com.google.common.collect.Maps;
import com.sk89q.jnbt.*;
import com.sk89q.worldedit.*;
import com.sk89q.worldedit.Vector;
import com.sk89q.worldedit.blocks.BaseBlock;
import com.sk89q.worldedit.bukkit.BukkitWorld;
import com.sk89q.worldedit.extension.input.ParserContext;
import com.sk89q.worldedit.extension.platform.Actor;
import com.sk89q.worldedit.extent.clipboard.BlockArrayClipboard;
import com.sk89q.worldedit.extent.clipboard.Clipboard;
import com.sk89q.worldedit.extent.clipboard.io.ClipboardFormat;
import com.sk89q.worldedit.extent.clipboard.io.ClipboardReader;
import com.sk89q.worldedit.extent.clipboard.io.SchematicReader;
import com.sk89q.worldedit.regions.CuboidRegion;
import com.sk89q.worldedit.session.ClipboardHolder;
import com.sk89q.worldedit.world.registry.WorldData;
import de.steamwar.sql.NoClipboardException;
import org.bukkit.entity.Player;

import java.io.*;
import java.util.*;
import java.util.logging.Level;
import java.util.stream.Collectors;

public class WorldEditWrapper8 implements WorldEditWrapper.IWorldEditWrapper {

    @Override
    public InputStream getPlayerClipboard(Player player, boolean schemFormat) {
        ClipboardHolder clipboardHolder;
        try {
            clipboardHolder = WorldEditWrapper.getWorldEditPlugin().getSession(player).getClipboard();
        } catch (EmptyClipboardException e) {
            throw new NoClipboardException();
        }

        Clipboard clipboard = clipboardHolder.getClipboard();
        if(clipboard == null)
            throw new NoClipboardException();

        PipedOutputStream outputStream = new PipedOutputStream();
        PipedInputStream inputStream;
        try {
            inputStream = new PipedInputStream(outputStream, 4096);
        } catch (IOException e) {
            throw new SecurityException("Could not init piped input stream", e);
        }

        new Thread(() -> {
            try {
                ClipboardFormat.SCHEMATIC.getWriter(outputStream).write(clipboard, clipboardHolder.getWorldData());
            } catch (IOException e) {
                Core.getInstance().getLogger().log(Level.SEVERE, "Could not write schematic", e);
            }
            try {
                outputStream.close();
            } catch (IOException e) {
                Core.getInstance().getLogger().log(Level.SEVERE, "Could not close schem writer", e);
            }
        }, "SchemWriter").start();

        return inputStream;
    }

    @Override
    public void setPlayerClipboard(Player player, InputStream is, boolean schemFormat) {
        WorldData world = new BukkitWorld(player.getWorld()).getWorldData();
        Clipboard clipboard;
        try {
            clipboard = getClipboard(is, schemFormat);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        Actor actor = WorldEditWrapper.getWorldEditPlugin().wrapCommandSender(player);
        WorldEditWrapper.getWorldEditPlugin().getWorldEdit().getSessionManager().get(actor).setClipboard(new ClipboardHolder(clipboard, world));
    }

    @Override
    public Clipboard getClipboard(InputStream is, boolean schemFormat) throws IOException {
        if(schemFormat)
            return new SpongeSchematicReader(new NBTInputStream(is)).read(WorldEdit.getInstance().getServer().getWorlds().get(0).getWorldData());
        else
            return new SchematicReader(new NBTInputStream(is)).read(WorldEdit.getInstance().getServer().getWorlds().get(0).getWorldData());
    }

    private static class SpongeSchematicReader implements ClipboardReader {

        private final NBTInputStream inputStream;
        private int schematicVersion = -1;

        SpongeSchematicReader(NBTInputStream inputStream) {
            Preconditions.checkNotNull(inputStream);
            this.inputStream = inputStream;
        }

        @Override
        public Clipboard read(WorldData worldData) throws IOException {
            CompoundTag schematicTag = this.getBaseTag();
            if (this.schematicVersion == 1) {
                return this.readSchematic(schematicTag);
            } else if (this.schematicVersion == 2) {
                return this.readSchematic(schematicTag);
            } else {
                throw new IOException("This schematic version is currently not supported");
            }
        }

        private CompoundTag getBaseTag() throws IOException {
            NamedTag rootTag = this.inputStream.readNamedTag();
            if (!rootTag.getName().equals("Schematic")) {
                throw new IOException("Tag 'Schematic' does not exist or is not first");
            } else {
                CompoundTag schematicTag = (CompoundTag)rootTag.getTag();
                Map<String, Tag> schematic = schematicTag.getValue();
                this.schematicVersion = (requireTag(schematic, "Version", IntTag.class)).getValue();
                return schematicTag;
            }
        }

        private BlockArrayClipboard readSchematic(CompoundTag schematicTag) throws IOException {
            IDConverter8 ids = new IDConverter8();

            Map<String, Tag> schematic = schematicTag.getValue();
            int width = (requireTag(schematic, "Width", ShortTag.class)).getValue();
            int height = (requireTag(schematic, "Height", ShortTag.class)).getValue();
            int length = (requireTag(schematic, "Length", ShortTag.class)).getValue();
            IntArrayTag offsetTag = getTag(schematic, "Offset", IntArrayTag.class);
            int[] offsetParts;
            if (offsetTag != null) {
                offsetParts = offsetTag.getValue();
                if (offsetParts.length != 3)
                    throw new IOException("Invalid offset specified in schematic.");
            } else {
                offsetParts = new int[]{0, 0, 0};
            }

            BlockVector min = new BlockVector(offsetParts[0], offsetParts[1], offsetParts[2]);
            CompoundTag metadataTag = getTag(schematic, "Metadata", CompoundTag.class);
            Vector origin;
            CuboidRegion region;
            if (metadataTag != null && metadataTag.containsKey("WEOffsetX")) {
                Map<String, Tag> metadata = metadataTag.getValue();
                int offsetX = (requireTag(metadata, "WEOffsetX", IntTag.class)).getValue();
                int offsetY = (requireTag(metadata, "WEOffsetY", IntTag.class)).getValue();
                int offsetZ = (requireTag(metadata, "WEOffsetZ", IntTag.class)).getValue();
                BlockVector offset = new BlockVector(offsetX, offsetY, offsetZ);
                origin = min.subtract(offset);
                region = new CuboidRegion(min, min.add(width, height, length).subtract(BlockVector.ONE));
            } else {
                origin = min;
                region = new CuboidRegion(min, min.add(width, height, length).subtract(BlockVector.ONE));
            }

            IntTag paletteMaxTag = getTag(schematic, "PaletteMax", IntTag.class);
            Map<String, Tag> paletteObject = requireTag(schematic, "Palette", CompoundTag.class).getValue();
            if (paletteMaxTag != null && paletteObject.size() != paletteMaxTag.getValue())
                throw new IOException("Block palette size does not match expected size.");

            Map<Integer, BaseBlock> palette = new HashMap<>();
            ParserContext parserContext = new ParserContext();
            parserContext.setRestricted(false);
            parserContext.setPreferringWildcard(false);

            for(String palettePart : paletteObject.keySet()) {
                IDConverter8.BlockTypeID blockID = ids.getId(palettePart);
                palette.put(requireTag(paletteObject, palettePart, IntTag.class).getValue(), new BaseBlock(blockID.getBlockId(), blockID.getDataId()));
            }

            byte[] blocks = requireTag(schematic, "BlockData", ByteArrayTag.class).getValue();
            Map<BlockVector, Map<String, Tag>> tileEntitiesMap = new HashMap<>();
            ListTag tileEntities = getTag(schematic, "BlockEntities", ListTag.class);
            if (tileEntities == null) {
                tileEntities = getTag(schematic, "TileEntities", ListTag.class);
            }

            if (tileEntities != null) {
                List<Map<String, Tag>> tileEntityTags = tileEntities.getValue().stream().map((tag) ->
                        (CompoundTag)tag
                ).map(CompoundTag::getValue).collect(Collectors.toList());

                BlockVector pt;
                Map<String, Tag> tileEntity;
                for(Iterator<Map<String, Tag>> var20 = tileEntityTags.iterator(); var20.hasNext(); tileEntitiesMap.put(pt, tileEntity)) {
                    tileEntity = var20.next();
                    int[] pos = requireTag(tileEntity, "Pos", IntArrayTag.class).getValue();
                    pt = new BlockVector(pos[0], pos[1], pos[2]);
                    Map<String, Tag> values = Maps.newHashMap(tileEntity);
                    values.put("x", new IntTag(pt.getBlockX()));
                    values.put("y", new IntTag(pt.getBlockY()));
                    values.put("z", new IntTag(pt.getBlockZ()));
                    values.put("id", values.get("Id"));
                    values.remove("Id");
                    values.remove("Pos");
                    tileEntity = values;
                }
            }

            BlockArrayClipboard clipboard = new BlockArrayClipboard(region);
            clipboard.setOrigin(origin);
            int index = 0;

            for(int i = 0; i < blocks.length; ++index) {
                int value = 0;
                int varintLength = 0;

                while(true) {
                    value |= (blocks[i] & 127) << varintLength++ * 7;
                    if (varintLength > 5) {
                        throw new IOException("VarInt too big (probably corrupted data)");
                    }

                    if ((blocks[i] & 128) != 128) {
                        ++i;
                        int y = index / (width * length);
                        int z = index % (width * length) / width;
                        int x = index % (width * length) % width;
                        BaseBlock block = palette.get(value);
                        BlockVector pt = new BlockVector(x, y, z);

                        try {
                            if (tileEntitiesMap.containsKey(pt)) {
                                block.setNbtData(new CompoundTag(tileEntitiesMap.get(pt)));
                                clipboard.setBlock(clipboard.getMinimumPoint().add(pt), block);
                            } else {
                                clipboard.setBlock(clipboard.getMinimumPoint().add(pt), block);
                            }
                            break;
                        } catch (WorldEditException var30) {
                            throw new IOException("Failed to load a block in the schematic");
                        }
                    }

                    ++i;
                }
            }

            return clipboard;
        }

        private static <T extends Tag> T requireTag(Map<String, Tag> items, String key, Class<T> expected) throws IOException {
            if (!items.containsKey(key)) {
                throw new IOException("Schematic file is missing a \"" + key + "\" tag");
            } else {
                Tag tag = items.get(key);
                if (!expected.isInstance(tag)) {
                    throw new IOException(key + " tag is not of tag type " + expected.getName());
                } else {
                    return expected.cast(tag);
                }
            }
        }

        private static <T extends Tag> T getTag(Map<String, Tag> items, String key, Class<T> expected) {
            if (!items.containsKey(key)) {
                return null;
            } else {
                Tag test = items.get(key);
                return !expected.isInstance(test) ? null : expected.cast(test);
            }
        }
    }
}
